#
# Copyright 2010 Andrey Churin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import sys
import tempfile
import shutil
from itertools import combinations, takewhile, izip

from closurebuilder.util import ClosureCommand, Command, Error, validate_name
from closurebuilder.metadata import Metadata
from closurebuilder.sourcemeta import Application
from closurebuilder.sources import Source
from closurebuilder.manifest import read_manifest
from closurebuilder.csscompiler import compile_css


__all__ = ['ClosureCompileCommand']


class ClosureCompileCommand(ClosureCommand):
    """Compile application"""
    summary = __doc__

    min_args = 1
    max_args = 1

    default_verbosity = 0

    group_name = 'Closure'

    parser = Command.standard_parser(simulate=True)

    def command_execute(self):
        "Main command to create client application"

        meta = Metadata.get_instance()
        meta.configure()

        dirname = os.path.join(meta.closure_dir, self.args[0])
        path = os.path.join(dirname, 'manifest.txt')
        meta.sources.include_strictly(dirname)
        app = read_manifest(path, processor = Application(dirname))

        print 'Searching closure library base...'
        base_file = meta.sources.get_base_file()

        resource_file = tempfile.NamedTemporaryFile(suffix='.js')
        resource_file.file.write('/** @param {string} cssFile*/goog.resource='
                                 'function(cssFile) {};\n')
        resource_file.file.flush()
        self.base_files = [base_file, Source(resource_file.name)]
        self.compile_app(app)

    def calculate_dependencies(self, app):
        def get_combined_paths(a, b):
            result = []
            for i in [map(lambda x: x[0],
                takewhile(lambda x: x[0] == x[1],
                    izip(i, j))) for i in a for j in b]:
                if not i in result:
                    result.append(i)
            return result

        print 'Calculating dependencies...'

        while 1:
            intersections = {}
            for i in app.modules:
                for j in i.get_sorted_sources():
                    intersections[j] = intersections.setdefault(j, []) + [i]

            intersections = dict(filter(lambda x: len(x[1]) > 1,
                                            intersections.items()))

            if not intersections:
                break

            for source, modules in intersections.iteritems():
                combined_paths = modules[0].get_dep_paths()
                for i in modules[1:]:
                    combined_paths = get_combined_paths(combined_paths,
                                                                i.get_dep_paths())
                # include source into the last module of
                # the lonest combined path
                max(combined_paths)[-1].add_include(source)

            for i in [app] + app.modules:
                provides = set()
                for j in i._include:
                    provides.update(j.provides)
                if not provides:
                    continue
                if len(provides) > 1:
                    s = 'These namespaces have'
                else:
                    s = 'This namespace has'
                print '%s been moved to %s:\n  %s'% \
                        (s, i.entry_point, ', '.join(provides))

    def compile_app(self, app):
        def print_namespaces(prefix, namespaces, resources):
            lines = [[]]
            for i in namespaces:
                if len(', '.join(lines[-1] + [i])) >= 78:
                    lines.append([])
                lines[-1].append(i)

            if len(namespaces)>1:
                this_namespace = 'these namespaces:'
            else:
                this_namespace = 'this namespace:'

            print prefix, 'includes', this_namespace
            print '  ' + '\n  '.join(map(', '.join, lines))

            if resources:
                print prefix.split(' ', 1)[0], 'resources:'
                for i in resources:
                    print '  ' + i

        def makedirs(path):
            if not os.path.isdir(path):
                os.makedirs(path)

        def get_resources(app):
            files = app.get_resources()

            files_set = set(files)
            for i in app.modules:
                module_files = set(i.get_resources())
                more_than_once = files_set.intersection(module_files)
                if more_than_once:
                    raise Error('These resource files was added more then '
                                'once:\n%s'%'\n'.join(more_than_once))
                files_set.update(module_files)

            resources = {}

            if app.get_css_filename():
                for i in app.modules:
                    files.extend(i.get_resources())
            else:
                for i in app.modules:
                    module_files = i.get_resources()
                    if module_files:
                        resources[i] = module_files
  
            if files:
                resources[app] = files

            return resources


        print '-'*80
        print 'Compiling "%s"'%app.entry_point
        print '-'*80

        self.calculate_dependencies(app)

        css_map = {}
        app.temp_dir, app.temp_path = os.path.split(tempfile.mktemp())

        print_namespaces('Application module', app.get_provides(),
                                                        app.get_resources())

        for i in app.modules:
            sources = i.get_sorted_sources()
            if not sources:
                raise Error('The module "%s" does not have any source files.'%
                    i.entry_point)
            print_namespaces('Module "%s"'%i.entry_point, i.get_provides(),
                i.get_resources())
            i.temp_path = os.path.split(tempfile.mktemp())[1]

        print 'Compiling resources...'
        
        resources = get_resources(app)

        for i in resources:
          resources[i] = compile_css(resources[i], css_map)

        base_files = list(self.base_files)

        if css_map:
            map_file = tempfile.NamedTemporaryFile(suffix='.js')
            map_content = 'goog.setCssNameMapping({%s});\n' % ',\n'.join(
                map(lambda i: "'%s':'%s'"%i, css_map.items()))
            map_file.write(map_content)
            map_file.flush()
            base_files.append(Source(map_file.name))

        if app.modules:
            module_file = tempfile.NamedTemporaryFile(suffix='.js')
            def module_info(module):
                return '"%s": [%s]'%(module.entry_point,
                            ','.join(map(lambda x: '"%s"'%x.entry_point,
                                   filter(lambda x: x != app,
                                        module.get_depends_on()))))
            info = ','.join(map(module_info, app.modules))
            uris = ','.join(map(lambda x:'"%s":["%s"]'%(x.entry_point, x.uri),
                                app.modules))
            meta = []
            for i in app.modules:
                m = i.get_meta()
                if m:
                    meta.append('"%s":%s'%(i.entry_point, i.get_meta()))
            meta = ','.join(meta)
            module_file.write("""
/**
 * @type {Object}
 * @private
 */
goog.module.ModuleInfo.prototype.meta_;

/**
 * @return {Object}
 */
goog.module.ModuleInfo.prototype.getMeta = function() {
  return this.meta_
}

/**
 * @param {Object} meta
 */
goog.module.ModuleInfo.prototype.setMeta = function(meta) {
  this.meta_ = meta;
}

/**
 * @param {Function} f
 * @param {Object=} opt_obj
 */
goog.module.ModuleManager.prototype.forEachModule = function(f, opt_obj) {
  for (var id in this.moduleInfoMap_) {
    f.call(opt_obj, this.moduleInfoMap_[id], id);
  }
}


/**
 * @param {Object.<string,Object>} moduleMetaMap
 */
goog.module.ModuleManager.prototype.setModuleMeta = function(moduleMetaMap) {
  for (var id in moduleMetaMap) {
    this.moduleInfoMap_[id].setMeta(moduleMetaMap[id]);
  }
}


var __moduleManager = goog.module.ModuleManager.getInstance();
__moduleManager.setLoader(new goog.module.ModuleLoader());
__moduleManager.setAllModuleInfo({%s});
__moduleManager.setModuleUris({%s});
__moduleManager.setModuleMeta({%s});
"""%(info, uris, meta))

            module_file.flush()
            module_source = Source(module_file.name)
        else:
            module_source = None

        compiler = Metadata.get_instance().compiler
        result, errout = compiler.compile(app, base_files, module_source, 
                                                          verbose=self.verbose)

        if result != 0:
            raise Error(errout)
        else:
            print errout

        makedirs(app.target)
        temp_dir = app.temp_dir
        for i in [app] + app.modules:
            src = os.path.join(temp_dir, i.temp_path)
            dst, ext = os.path.splitext(os.path.join(app.target,
                                                        i.get_module_name()))
            if not ext:
                js_dst = dst + '.js'
            else:
                js_dst = dst + ext

            print 'Writing "%s"' % os.path.normpath(js_dst)
            makedirs(os.path.dirname(js_dst))
            shutil.move(src + '.js', js_dst)

            if not app.get_css_filename() and i in resources:
                print 'Writing "%s"' % (os.path.normpath(dst) + '.css')
                f = open(dst + '.css', 'wb')
                f.write(resources[i])
                f.close()

        if app.get_css_filename() and app in resources:
            target = os.path.join(app.target, app.get_css_filename())
            makedirs(os.path.dirname(target))
            print 'Writing "%s"' % (os.path.normpath(target))
            f = open(target, 'wb')
            f.write(resources[app])
            f.close()

        print 'Compilation succeeded'
