#
# Copyright 2010 Andrey Churin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
import os
import distutils.version
import subprocess
import pkg_resources

from closurebuilder.util import squeeze, Error
from closurebuilder.manifest import manifest_directive

_VERSION_REGEX = re.compile('[\.0-9]+')

class Compiler(object):
    "compiler-settings.txt processor"

    ERRORS = ('accessControls', 'checkRegExp', 'checkTypes', 'checkVars',
              'deprecated', 'fileoverviewTags', 'invalidCasts',
              'missingProperties', 'nonStandardJsDocs', 'strictModuleDepCheck',
              'undefinedVars', 'unknownDefines', 'visibility')

    def __init__(self, path):
        self._path = path
        self._charset = 'UTF-8'
        self._comp_level = 'ADVANCED_OPTIMIZATIONS'
        self._jscomp_off = []
        self._jscomp_error = list(self.ERRORS)
        self._jscomp_warning = []
        self._warning_level = 'VERBOSE'
        self._summary_detail_level = '3'
        self._compiler = None

    def _add_comp(self, option, value):
        if value in self._jscomp_off:
          self._jscomp_off.remove(value)
        elif value in self._jscomp_error:
          self._jscomp_error.remove(value)
        else:
          self._jscomp_warning.remove(value)
        getattr(self, '_' + option).append(value)

    @manifest_directive
    def compiler(self, opt, value):
        "compiler %"
        path = os.path.join(self._path, value)
        if not os.path.isfile(path):
            raise Error('File "%s" does not exists'%path, 9)
        self._compiler = path
        
    @manifest_directive
    def charset(self, opt, value):
        "charset %"
        self._charset = value

    @manifest_directive
    def jscomp_error(self, opt, value):
        "enable error %"
        if not value in self.ERRORS:
            raise Error('Invalid error class', 13)
        self._add_comp('jscomp_error', value)

    @manifest_directive
    def jscomp_warning(self, opt, value):
        "enable warning %"
        if not value in self.ERRORS:
            raise Error('Invalid error class', 15)
        self._add_comp('jscomp_warning', value)

    @manifest_directive
    def jscomp_off(self, opt, value):
        "disable warning %"
        if not value in self.ERRORS:
            raise Error('Invalid error class', 16)
        self._add_comp('jscomp_off', value)

    @manifest_directive
    def summary_detail_level(self, opt, value):
        "summary detail level %"
        if not value in ('0','1','2','3'):
            raise Error('Invalid detail level', 21)
        self._summary_detail_level = value

#    @manifest_directive
#    def use_only_custom_externs(self, opt):
#        "use only custom externs"
#        self._commands.append('--use_only_custom_externs')

    @manifest_directive
    def warning_level(self, opt, value):
        "warning level %"
        if not value in ('QUIET', 'DEFAULT', 'VERBOSE'):
            raise Error('Invalid warning level', 1)
        self._warning_level = value

    def _get_java_version(self):
        "Returns the string for the current version of Java installed."
        try:
            proc = subprocess.Popen(['java', '-version'], stderr=subprocess.PIPE)
            unused_stdoutdata, stderrdata = proc.communicate()
            version_line = stderrdata.splitlines()[0]
        except:
            return '0.0'
        return _VERSION_REGEX.search(version_line).group()

    def compile(self, app, extra_files=[], module_source=None, verbose=False):
        if not (distutils.version.LooseVersion(self._get_java_version()) >
            distutils.version.LooseVersion('1.6')):
            raise Error('Closure compiler requires Java 1.6 or higher. '
                        'Please visit http://www.java.com/getjava')

        if self._compiler:
          compiler_jar_path = os.path.abspath(self._compiler)
        else:
          compiler_jar_path = pkg_resources.resource_filename('closurebuilder',
                                                        'jar/compiler.jar')
        args = ['java', '-jar', compiler_jar_path]

        args += ['--charset', self._charset]
        args += ['--compilation_level', self._comp_level]
        args += ['--warning_level', self._warning_level]
        args += ['--summary_detail_level', self._summary_detail_level]
        for i in self._jscomp_off:
          args += ['--jscomp_off', i]
        for i in self._jscomp_warning:
          args += ['--jscomp_warning', i]
        for i in self._jscomp_error:
          args += ['--jscomp_error', i]
        args += app.get_compiler_settings()

        for i in extra_files:
            args += '--js', i.get_path()

        app_sources = app.get_sorted_sources()

        if module_source:
            placeTo = 0
            for i, j in enumerate(app_sources):
                if 'goog.module.ModuleManager' in j.provides or \
                       'goog.module.ModuleLoader' in j.provides:
                   placeTo = i
            app_sources.insert(i, module_source)

        for i in app_sources:
            args += '--js', i.get_path()
        args += '--module', '%s:%s'%(app.temp_path,
                                            len(app_sources) + len(extra_files))
        args += '--module_output_path_prefix', app.temp_dir + os.path.sep
        depends_on = [app]

        modules = list(app.modules)
        while modules:
            next_phase_modules = []
            for i in modules:
                if not set(depends_on).issuperset(i.get_depends_on()):
                    next_phase_modules.append(i)
                    continue
                sources = i.get_sorted_sources()
                for j in sources:
                    args += '--js', j.get_path()
                args += '--module', '%s:%s:%s'%(i.temp_path, len(sources),
                    i.get_depends_on()[-1].temp_path)
                depends_on.append(i)
            modules = next_phase_modules

        if verbose:
          print 'Compiling javascript with the following command:'
          print ' '.join(args)
        else:
          print 'Compiling javascript'

        proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE)
        stdoutdata, stderrdata = proc.communicate()

        return proc.returncode, stderrdata
